@@ -203,7 +203,7 @@ class IMPOConfig(TypedDict):
     stm_diff_coef: Optional[float]
     use_stm_cosine_sim: Optional[bool]
     cosine_sim_coef: Optional[float]
-    advantage_mode: Literal['batch', 'step']
+    advantage_mode: Literal['batch', 'step', 'all']
 
 
 class IMPOAlgorithm(RlAlgorithm):
@@ -242,7 +242,7 @@ class IMPOAlgorithm(RlAlgorithm):
         self.use_stm_cosine_sim = config.get('use_stm_cosine_sim', False)
         self.cosine_sim_coef = config.get('cosine_sim_coef', 0.01)
         self.debug_step = 0
-        self.advantage_mode = config.get('advantage_mode', 'batch')
+        self.advantage_mode = config.get('advantage_mode', 'all')
 
         # Additional losses
         self.policy_consistency_loss_fn = nn.KLDivLoss(reduction='batchmean', log_target=False)
@@ -319,6 +319,9 @@ class IMPOAlgorithm(RlAlgorithm):
         entropy = -(
             (new_log_probs * new_log_probs.exp() * entropy_mask.unsqueeze(-1)).sum(dim=-1) / (entropy_mask.sum(dim=-1).unsqueeze(-1) + 1e-8)
         ).mean()
+        if self.debug_step != 0 and self.debug_step % self.debug_interval == 0:
+            print(
+                f'---- Entropy bonus: {entropy.item():.4f}, scaled: {(self.entropy_coef * entropy).item():.4f}')
         policy_loss -= self.entropy_coef * entropy
 
         # 9. Reference KL-div penalty
@@ -352,7 +355,7 @@ class IMPOAlgorithm(RlAlgorithm):
 
     def kl_penalty(self, ref_log_probs: torch.Tensor, new_log_probs: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
         kl = (ref_log_probs.exp() * (ref_log_probs - new_log_probs)) * mask
-        kl_loss = kl.sum() / (mask.sum() + 1e-8)
+        kl_loss = (kl.sum(dim=-1) / (mask.sum(dim=-1) + 1e-8)).mean()
         return kl_loss
 
     def policy_consistency_loss(self, prev_step_log_probs: torch.Tensor, this_step_probs: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
@@ -396,17 +399,25 @@ class IMPOAlgorithm(RlAlgorithm):
         else:
             if self.use_gae:
                 normalize_mask = ~dones.unsqueeze(-1)[:-1, :].expand_as(advantages)
+                if self.advantage_mode == 'all':
+                    normalized_advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)
+                elif self.advantage_mode == 'batch':
+                    normalized_advantages = self.normalize_advantages_flatten(advantages, normalize_mask)
+                else:
+                    normalized_advantages = self.normalize_advantages_per_step_with_mask(advantages, normalize_mask)
             else:
-                normalize_mask = ~dones.unsqueeze(-1).expand_as(advantages)
+                if self.advantage_mode == 'all':
+                    normalized_advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-8)
+                elif self.advantage_mode == 'batch':
+                    normalize_mask = ~dones.unsqueeze(-1).expand_as(advantages)
+                    normalized_advantages = self.normalize_advantages_flatten(advantages, normalize_mask)
+                else:
+                    normalized_advantages = self.normalize_advantages_per_step(advantages)
 
-            if self.advantage_mode == 'batch':
-                normalized_advantages = self.normalize_advantages_flatten(advantages, normalize_mask)
-            else:
-                normalized_advantages = self.normalize_advantages_per_step(advantages, normalize_mask)
         return normalized_advantages, ref_values
 
-    def normalize_advantages_per_step(self, advantages: torch.Tensor, mask: torch.Tensor, eps=1e-8,
-                                      clip=10.0) -> torch.Tensor:
+    def normalize_advantages_per_step_with_mask(self, advantages: torch.Tensor, mask: torch.Tensor, eps: float = 1e-8,
+                                      clip: float = 10.0) -> torch.Tensor:
         """
         advantages: [T, B]
         mask: [T, B]
@@ -431,6 +442,26 @@ class IMPOAlgorithm(RlAlgorithm):
         out = out.transpose(0, 1).contiguous()
         return out
 
+    def normalize_advantages_per_step(self, advantages: torch.Tensor, eps: float = 1e-8,
+                                      clip: float = 10.0) -> torch.Tensor:
+        """
+        advantages: [T, B]
+        mask: [T, B]
+        Normalizes along batch axis for each timestep t separately.
+        """
+
+        T, B = advantages.shape
+        out = torch.zeros_like(advantages)
+        for t in range(T):
+            vals = advantages[t, :]
+            mean = vals.mean()
+            std = vals.std(unbiased=False)
+            normed = (vals - mean) / (std + eps)
+            normed = torch.clamp(normed, -clip, clip)
+            out[t, :] = normed
+
+        return out
+
     def normalize_advantages_flatten(self, advantages: torch.Tensor, mask: torch.Tensor, eps=1e-8, clip=10.0) -> torch.Tensor:
         """
         advantages: [T, B]  (we'll handle both)
